unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ExtCtrls, StdCtrls,
  Grids, typ;

type
  TArbFloatPoint = record x, y: ArbFloat; end;
  T4Points = array[0..3] of TArbFloatPoint;
  TCoeffs = array[0..7] of ArbFloat;

  { TForm1 }

  TForm1 = class(TForm)
    Button1: TButton;
    Image1: TImage;
    Image2: TImage;
    Label1: TLabel;
    Grid: TStringGrid;
    procedure Button1Click(Sender: TObject);
    procedure Image1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    PO: T4Points;

  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

uses
  sle, FPImage, IntfGraphics;

const
  PD: T4Points = (          // Ziel-Bild ("D" = destination)
    (x:  0;   y:    0),
    (x:  0;   y:290*2),
    (x:210*2; y:290*2),
    (x:210*2; y:    0)
  );

// xO = (c0 * xD + c1 * yD + c2) / (c6 * xD + c7 * yD +1)
// yO = (c3 * xD + c4 * yD + c5) / (c6 * xD + c7 * yD +1)
// Quellpunkte P, Zielpunkte Q
function CalcCoeffs(const P, Q: T4Points; out c:TCoeffs): Integer;
const
  N = 8;
var
  a: array[0..N-1, 0..N-1] of ArbFloat;
  b: array[0..N-1] of ArbFloat;
  i: Integer;
  j: Integer;
  ca: ArbFloat;
  term: ArbInt;
begin
  // Matrix des Gleichungssystems
  j := 0;
  for i := 0 to 3 do begin
    a[j, 0] := Q[i].X;
    a[j, 1] := Q[i].Y;
    a[j, 2] := 1.0;
    a[j, 3] := 0.0;
    a[j, 4] := 0.0;
    a[j, 5] := 0.0;
    a[j, 6] := -P[i].X * Q[i].X;
    a[j, 7] := -P[i].X * Q[i].Y;
    inc(j);

    a[j, 0] := 0.0;
    a[j, 1] := 0.0;
    a[j, 2] := 0.0;
    a[j, 3] := Q[i].X;
    a[j, 4] := Q[i].Y;
    a[j, 5] := 1.0;
    a[j, 6] := -P[i].Y * Q[i].X;
    a[j, 7] := -P[i].Y * Q[i].Y;
    inc(j);
  end;

  // Vektor für rechte Seite
  for j := 0 to 3 do
  begin
    b[j*2] := P[j].X;
    b[j*2+1] := P[j].Y;
  end;

  // Gleichungssystem lösen
  slegen(n, n, a[0, 0], b[0], c[0], ca, term);

  Result := term;
end;


{ TForm1 }

// https://en.wikipedia.org/wiki/Bilinear_interpolation#Unit_square
//          A ----- B
//           |   +  |             +: (u,v)  u,v = 0..1
//           |      |
//          C  ---- D
function BilinearInterpolation(u, v: Double; A,B,C,D: Integer): Integer;
begin
  Result := trunc(A*(1-u)*(1-v) + B*u*(1-v) + C*(1-u)*v + D*u*v);
end;

function BilinearColorInterpolation(x,y: Double; ClrA,ClrB,ClrC,ClrD: TFPColor): TFPColor;
var
  u, v: Double;
begin
  u := frac(x);
  v := frac(y);
  Result.Red := BilinearInterpolation(u, v, ClrA.Red, ClrB.Red, ClrC.Red, ClrD.Red);
  Result.Green := BilinearInterpolation(u, v, ClrA.Green, ClrB.Green, ClrC.Green, ClrD.Green);
  Result.Blue := BilinearInterpolation(u, v, ClrA.Blue, ClrB.Blue, ClrC.Blue, ClrD.Blue);
end;

procedure TForm1.Button1Click(Sender: TObject);
const
  ERR_MSG = 'Alle Eckpunkte des zu entzerrenden Rechtecks müssen komplett festgelegt sein.';
var
  i: Integer;
  c: TCoeffs;
  x, y: Integer;
  u, v: Double;
  value: double;
  denom: Double;
  clrA, clrB, clrC, clrD: TFPColor;
  srcImg, destImg: TLazIntfImage;
  res: Integer;
begin
  for i := 0 to 3 do begin
    if tryStrToFloat(Grid.Cells[1, i+1], value) then
      PO[i].X := value
    else begin
      MessageDlg(ERR_MSG, mtError, [mbOK], 0);
      exit;
    end;
    if TryStrToFloat(Grid.Cells[2, i+1], value) then
      PO[i].Y := value
    else begin
      MessageDlg(ERR_MSG, mtError, [mbOK], 0);
      exit;
    end;
  end;

  Screen.Cursor := crHourglass;
  try
    res := CalcCoeffs(PO, PD, c);
    srcImg := Image1.Picture.Bitmap.CreateIntfImage;
    try
      destImg := TLazIntfImage.Create(0, 0);
      try
        destImg.DataDescription := srcImg.DataDescription;
        destImg.SetSize(round(PD[2].X), round(PD[2].Y));
        for y := 0 to destImg.Height - 1 do
        begin
          for x := 0 to destImg.Width - 1 do
          begin
            denom := (c[6] * x + c[7] * y + 1);
            u := (c[0] * x + c[1] * y + c[2]) / denom;
            v := (c[3] * x + c[4] * y + c[5]) / denom;
            clrA := srcImg.Colors[trunc(u), trunc(v)];
            clrB := srcImg.Colors[trunc(u)+1, trunc(v)];
            clrC := srcImg.Colors[trunc(u), trunc(v+1)];
            clrD := srcImg.Colors[trunc(u+1), trunc(v+1)];
            destImg.Colors[x, y] := BilinearColorInterpolation(u, v, clrA, clrB, clrC, clrd);
          end;
        end;
        Image2.Width := destImg.Width;
        Image2.Height := destImg.Height;
        Image2.Picture.Bitmap.LoadFromIntfImage(destImg);
      finally
        destImg.Free;
      end;
    finally
      srcImg.Free;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;

procedure TForm1.Image1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  Grid.Cells[1, Grid.Row] := IntToStr(X);
  Grid.Cells[2, Grid.Row] := IntToStr(Y);
  if Grid.Row < Grid.RowCount-1 then
    Grid.Row := Grid.Row + 1;
end;

end.

